import numpy as np
from scipy.stats import ttest_ind, norm
from django.db import models
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth import models as auth_models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.core.exceptions import ValidationError
from abc import ABC, abstractmethod

from waffle.models import Flag  # as WaffleFlag
from waffle.models import AbstractUserFlag

DEBUG = settings.DEBUG if hasattr(settings, "DEBUG") else False

BANDIT_ALGORITHMS = [
    ("EG", "Epsilon Greedy"),
    ("ED", "Epsilon Decay"),
    ("UCB1", "Upper Confidence Bound"),
]


class BanditFlag(AbstractUserFlag):
    # TODO: Are content_type and object_id necessary?
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, blank=True
    )
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = GenericForeignKey("content_type", "object_id")

    def is_active_for_user(self, user):
        # TODO: Update to reflect the new bandit model
        User = get_user_model()
        if not user:
            try:
                user = User.objects.get(
                    username="username"
                )  # replace 'username' with actual username
            except User.DoesNotExist:
                user = auth_models.AnonymousUser()

        # Try to retrieve the active bandit model
        bandit_model_instance = None
        for related_set in [
            self.epsilongreedymodel_set,
            self.epsilondecaymodel_set,
            self.ucb1model_set,
        ]:  # Add more sets as needed
            bandit_model_instance = related_set.filter(is_active=True).first()
            if bandit_model_instance:
                break

        if bandit_model_instance is None:
            return super().is_active_for_user(user)

        # Call pull method on the bandit model instance
        active = bandit_model_instance.pull()
        return active


class URLSanitizationMixin:
    url_field_names = ["url", "source_url", "target_url"]

    def save(self, *args, **kwargs):
        for field_name in self.url_field_names:
            url = getattr(self, field_name, None)
            if url is None:
                continue
            setattr(self, field_name, self.ensure_trailing_slash(url))
        super().save(*args, **kwargs)

    @staticmethod
    def ensure_trailing_slash(url):
        # Ensure it has a trailing and leading slash
        if not url.endswith("/"):
            url += "/"

        if not url.startswith("/"):
            url = "/" + url

        return url


class FlagUrl(URLSanitizationMixin, models.Model):
    flag = models.OneToOneField(BanditFlag, on_delete=models.CASCADE)
    source_url = models.CharField(null=True, max_length=200, blank=True)
    target_url = models.CharField(null=True, max_length=200, blank=True)
    active_flag_views = models.IntegerField(default=0)
    inactive_flag_views = models.IntegerField(default=0)
    active_flag_conversions = models.IntegerField(default=0)
    inactive_flag_conversions = models.IntegerField(default=0)


class UserActivity(URLSanitizationMixin, models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL
    )
    session_key = models.CharField(max_length=40, null=True)
    url = models.URLField()
    target_url_visit = models.BooleanField(default=False)
    timestamp = models.DateTimeField(auto_now_add=True)
    flags = models.ManyToManyField(BanditFlag, blank=True)
    is_staff = models.BooleanField(default=False)


class UserActivityFlag(models.Model):
    user_activity = models.ForeignKey(UserActivity, on_delete=models.CASCADE)
    flag = models.ForeignKey(BanditFlag, on_delete=models.CASCADE)
    timestamp = models.DateTimeField(auto_now_add=True)
    is_active = models.BooleanField(default=False, null=True)


class Bandit(models.Model):
    name = models.CharField(max_length=200, choices=BANDIT_ALGORITHMS, default="EG")
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")
    is_active = models.BooleanField(default=False)
    timestamp = models.DateTimeField(auto_now_add=True)

    def save(self, *args, **kwargs):
        """Ensure only one active bandit is available at a time."""
        if self.is_active and isinstance(self.content_object, BanditFlag):
            bandit_flag_content_type = ContentType.objects.get_for_model(BanditFlag)
            active_bandits = Bandit.objects.filter(
                content_type=bandit_flag_content_type,
                object_id=self.content_object.id,
                is_active=True,
            )
            if self.pk:
                active_bandits = active_bandits.exclude(pk=self.pk)
            if active_bandits.exists():
                raise ValidationError("Only one bandit can be active at a time.")
        super().save(*args, **kwargs)


# TODO: Deprecated, delete
class BanditInstance(models.Model):
    # Used for Admin purposes
    flag = models.ForeignKey(BanditFlag, on_delete=models.CASCADE)
    bandit = models.OneToOneField(Bandit, on_delete=models.CASCADE)


class AbstractBanditModel(models.Model):
    flag = models.ForeignKey(
        BanditFlag,
        related_name="%(class)s_set",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    is_active = models.BooleanField(default=False)
    significance_level = models.FloatField(default=0.05)
    min_views = models.IntegerField(default=100)
    winning_arm = models.IntegerField(null=True, blank=True)

    k = 2  # Number of options

    class Meta:
        abstract = True
        constraints = [
            models.UniqueConstraint(
                fields=["flag", "is_active"],
                condition=models.Q(is_active=True),
                name="unique_active_bandit_%(class)s",
            )
        ]

    # def __str__(self):
    #   return self.display_bounds()

    @abstractmethod
    def pull(self):
        """Determines whether or not the flag is active 0 = False, 1 = True"""
        raise NotImplementedError("Pull method not implemented.")

    def get_rewards(self):
        """Returns the rewards for each option"""
        rewards = self.get_number_of_conversions() / np.maximum(
            self.get_number_of_views(), 1
        )
        return rewards

    def get_number_of_views(self):
        """Returns the number of views for each option"""
        flag_url = self.flag.flagurl
        return np.array([flag_url.inactive_flag_views, flag_url.active_flag_views])

    def get_number_of_conversions(self):
        """Returns the number of conversions for each option"""
        flag_url = self.flag.flagurl
        return np.array(
            [flag_url.inactive_flag_conversions, flag_url.active_flag_conversions]
        )

    def save(self, *args, **kwargs):
        """Ensure only one bandit is active at a time."""
        if self.is_active:
            active_bandits = type(self).objects.filter(flag=self.flag, is_active=True)
            if self.pk:
                active_bandits = active_bandits.exclude(pk=self.pk)
            if active_bandits.exists():
                raise ValidationError("Only one bandit can be active at a time.")
        super().save(*args, **kwargs)

    def test_arms(self):
        """
        Performs a two-sample t-test on the rewards for each arm
        """
        n_views = self.get_number_of_views()
        if n_views.sum() < self.min_views:
            return None

        # Converts outcomes to list of binary outcomes for use with ttest_ind
        n_convs = self.get_number_of_conversions()
        inactive_results = [1] * n_convs[0] + [0] * (n_views[0] - n_convs[0])
        active_results = [1] * n_convs[1] + [0] * (n_views[1] - n_convs[1])

        t, p_value = ttest_ind(inactive_results, active_results)
        if p_value < self.significance_level:
            self.winning_arm = (
                0 if np.mean(inactive_results) > np.mean(active_results) else 1
            )
            self.save()

    def get_confidence_intervals(self, arm: int) -> tuple:
        """Gets upper and lower bounds for the given arm"""
        n_views = self.get_number_of_views()[arm]
        n_convs = self.get_number_of_conversions()[arm]
        alpha = 1 - self.significance_level
        prop = n_convs / n_views
        z = norm.ppf(1 - alpha / 2)
        error = z * np.sqrt(prop * (1 - prop) / n_views)
        return prop - error, prop + error

    def get_inactive_flag_confidence_intervals(self) -> str:
        low, up = self.get_confidence_intervals(0)
        return f"{low:.2f} - {up:.2f}"

    def get_active_flag_confidence_intervals(self) -> str:
        low, up = self.get_confidence_intervals(1)
        return f"{low:.2f} - {up:.2f}"

    def display_confidence_intervals(self):
        inactive = self.get_inactive_flag_confidence_intervals()
        active = self.get_active_flag_confidence_intervals()
        return f"Active Flag:\t{active}\nInactive Flag:\t{inactive}"

    def display_conversion_rate(self):
        rewards = self.get_rewards()
        return f"Active Flag:\t{rewards[1]:.2%}\nInactive Flag:\t{rewards[0]:.2%}"


class EpsilonGreedyModel(AbstractBanditModel):
    epsilon = models.FloatField(default=0.1)
    prob_flag = models.FloatField(default=0.5)

    def pull(self) -> bool:
        if self.winning_arm is not None:
            return bool(self.winning_arm)
        p = np.random.random()
        if p < self.epsilon:
            if DEBUG:
                print("Making random choice")
            flag = np.random.randint(0, self.k)
        else:
            rewards = self.get_rewards()
            if DEBUG:
                print("Making greedy choice")
                print(f"Rewards: {rewards}")
            if rewards[0] == rewards[1]:
                flag = np.random.randint(0, self.k)
            else:
                flag = np.argmax(rewards)

        return bool(flag)

    def update(self):
        """
        Updates probability of flag vs. no flag and any other stats

        TODO: Move this to be calculated on admin page
        """
        rewards = self.get_rewards()
        if rewards[0] == rewards[1]:
            self.prob_flag = 0.5
        else:
            flag = bool(np.argmax(rewards))
            if flag:
                self.prob_flag = 1 - self.epsilon / 2
            else:
                self.prob_flag = self.epsilon / 2


class EpsilonDecayModel(AbstractBanditModel):
    def pull(self) -> bool:
        if self.winning_arm is not None:
            return bool(self.winning_arm)
        p = np.random.random()
        if p < (1 / (1 + self.get_number_of_views().sum() / self.k)):
            flag = np.random.randint(0, self.k)
        else:
            rewards = self.get_rewards()
            if rewards[0] == rewards[1]:
                flag = np.random.randint(0, self.k)
            else:
                flag = np.argmax(rewards)

        return bool(flag)


class UCB1Model(AbstractBanditModel):
    c = models.FloatField(default=2.0)

    def pull(self) -> bool:
        """
        Pulls the arm with the highest upper confidence bound
        """
        if self.winning_arm is not None:
            return bool(self.winning_arm)
        rewards = self.get_rewards()
        n_views = self.get_number_of_views()
        flag = np.argmax(
            rewards
            + self.c
            * np.sqrt(np.log(np.max([n_views.sum(), 1])) / np.maximum(n_views, 1))
        )
        return bool(flag)



from django.contrib import admin
from .models import (
    UserActivity,
    FlagUrl,
    UserActivityFlag,
    BanditFlag,
    EpsilonGreedyModel,
    EpsilonDecayModel,
    UCB1Model,
)
from .forms import BanditAdminForm  # Deprecated? Delete?


class FlagUrlInline(admin.StackedInline):
    model = FlagUrl
    extra = 0


class BaseBanditInline(admin.TabularInline):
    extra = 0
    fk_name = "flag"
    readonly_fields = ["display_conversion_rate", "display_confidence_intervals"]
    fields = [
        "is_active",
        "significance_level",
        "min_views",
        "winning_arm",
        "display_conversion_rate",
        "display_confidence_intervals",
    ]
    new_fields = []
    new_field_positions = []

    def display_conversion_rate(self, obj):
        return obj.display_conversion_rate()

    display_conversion_rate.short_description = "Conversion Rate"

    def display_confidence_intervals(self, obj):
        return obj.display_confidence_intervals()

    display_confidence_intervals.short_description = "Confidence Intervals"

    def insert_bandit_fields(self, bandit_fields: list, positions: list) -> None:
        """For inserting bandit specific fields into the admin display"""
        new_fields = list(
            self.fields
        )  # Make a copy to avoid overwriting for all classes
        for i, f in zip(positions, bandit_fields):
            new_fields.insert(i, f)
        self.fields = new_fields

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.insert_bandit_fields(self.new_fields, self.new_field_positions)


class EpsilonDecayModelInline(BaseBanditInline):
    model = EpsilonDecayModel


class EpsilonGreedyModelInline(BaseBanditInline):
    model = EpsilonGreedyModel
    new_fields = ["epsilon"]
    new_field_positions = [1]


class UCB1ModelInline(BaseBanditInline):
    model = UCB1Model
    new_fields = ["c"]
    new_field_positions = [1]


# Define the admin interface for Flag
class FlagAdmin(admin.ModelAdmin):
    inlines = [
        FlagUrlInline,
        EpsilonGreedyModelInline,
        EpsilonDecayModelInline,
        UCB1ModelInline,
    ]


class UserActivityFlagInline(admin.TabularInline):
    model = UserActivityFlag
    extra = 1


class UserActivityInline(admin.TabularInline):
    model = UserActivity


class UserActivityAdmin(admin.ModelAdmin):
    inlines = [UserActivityFlagInline]
    fieldsets = [
        (
            None,
            {
                "fields": [
                    "user",
                    "session_key",
                    "url",
                    "is_staff",
                    "flags",
                    "target_url_visit",
                ]
            },
        )
    ]
    readonly_fields = ["timestamp"]
    list_display = ["user", "is_staff", "session_key", "url", "timestamp"]


# Unregister the original Flag admin form provided by Django Waffle
# admin.site.unregister(Flag)

# Register the new FlagAdmin
admin.site.register(BanditFlag, FlagAdmin)
admin.site.register(UserActivity, UserActivityAdmin)



from django import forms

from .models import Bandit


class BanditAdminForm(forms.ModelForm):
    class Meta:
        model = Bandit
        fields = "__all__"

    def clean(self):
        cleaned_data = super().clean()
        is_active = cleaned_data.get("is_active")
        flag = cleaned_data.get("flag")
        if is_active:
            active_bandits = Bandit.objects.filter(flag=flag, is_active=True)
            if self.instance.pk:
                active_bandits = active_bandits.exclude(pk=self.instance.pk)
            if active_bandits.exists():
                self.add_error(None, "Only one bandit can be active at a time.")



import re
from django.conf import settings
from django.http import HttpRequest, HttpResponse
from .models import UserActivity, FlagUrl, UserActivityFlag
from .models import BanditFlag

import waffle

DEBUG = settings.DEBUG if hasattr(settings, "DEBUG") else False


class UserActivityMiddleware:
    # TODO: Exclude 404 pages
    exclude = (
        settings.EXCLUDE_FROM_TRACKING
        if hasattr(settings, "EXCLUDE_FROM_TRACKING")
        else ["/admin"]
    )
    exclude = [path if path.startswith("/") else "/" + path for path in exclude]
    exclude_regex = (
        settings.EXCLUDE_FROM_TRACKING_REGEX
        if hasattr(settings, "EXCLUDE_FROM_TRACKING_REGEX")
        else ""
    )

    def __init__(self, get_response):
        self.get_response = get_response

    def flag_is_active(self, request, flag_name):
        """
        Returns whether or not a flag is active for a given request

        This is a wrapper around waffle.flag_is_active that allows us to mock
        for testing purposes.
        """
        return waffle.flag_is_active(request, flag_name)

    def check_exclusion(self, current_url):
        # Dynamic loading of settings for debugging
        if DEBUG:
            exclude = getattr(settings, "EXCLUDE_FROM_TRACKING", ["/admin"])
            self.exclude = [
                path if path.startswith("/") else "/" + path for path in exclude
            ]
            self.exclude_regex = getattr(settings, "EXCLUDE_FROM_TRACKING_REGEX", "")
        if any(current_url.startswith(path) for path in self.exclude):
            print(f"{current_url} excluded by exclusion path: {self.exclude}")
            return True
        if self.exclude_regex and re.search(self.exclude_regex, current_url):
            print(f"{current_url} excluded by regex: {self.exclude_regex}")
            return True
        return False

    def __call__(self, request: HttpRequest) -> HttpResponse:
        session_key = request.session.session_key
        current_url = request.path

        if self.check_exclusion(current_url):
            return self.get_response(request)

        # If the user is authenticated, add their user instance to the UserActivity
        if request.user.is_authenticated:
            user_activity = UserActivity.objects.create(
                user=request.user,
                is_staff=request.user.is_staff,
                session_key=session_key,
                url=current_url,
            )
        else:
            user_activity = UserActivity.objects.create(
                session_key=session_key, url=current_url
            )

        # Checks to see which flags are active on the source page, if any
        for flag in BanditFlag.objects.all():
            flag_url = FlagUrl.objects.filter(flag=flag).first()
            if flag_url:
                if current_url == flag_url.source_url:
                    # is_flag_active determines whether or not the user sees the feature
                    is_flag_active = self.flag_is_active(request, flag.name)
                    # if DEBUG:
                    #     print(
                    #         f"Checking flag {flag.name} for user {request.user}\nFlag URL: {flag_url.source_url}\nFlag is active: {is_flag_active}"
                    #     )
                    if is_flag_active is not None:
                        ua_flag = UserActivityFlag.objects.create(
                            user_activity=user_activity,
                            flag=flag,
                            is_active=is_flag_active,
                        )
                        ua_flag.save()
                        if is_flag_active:
                            flag_url.active_flag_views += 1
                        else:
                            flag_url.inactive_flag_views += 1
                        flag_url.save()

                # Checks to see if the user has reached the target URL
                elif current_url == flag_url.target_url:
                    source_reached = UserActivity.objects.filter(
                        session_key=session_key, url=flag_url.source_url
                    ).exists()
                    if source_reached:
                        user_activity.target_url_visit = True
                        user_activity.save()
                        # Need to see which source flag the user saw before, if any
                        active_source_flag = self._get_latest_flagged_source_url(
                            session_key, flag_url.source_url
                        )
                        # TODO: Update to handle cases where users didn't see the particular
                        # landing page or feature flag at all
                        # if DEBUG:
                        #     print(
                        #         f"User reached target URL {flag_url.target_url}\nActive source flag: {active_source_flag}"
                        #     )
                        if active_source_flag:
                            flag_url.active_flag_conversions += 1
                        else:
                            flag_url.inactive_flag_conversions += 1

                        flag_url.save()

                        # Update bandit stats
                        for related_set in [
                            flag.epsilongreedymodel_set,
                            flag.epsilondecaymodel_set,
                            flag.ucb1model_set,
                        ]:
                            bandit_model_instance = related_set.filter(
                                is_active=True
                            ).first()
                            if bandit_model_instance:
                                bandit_model_instance.test_arms()
                                break

        response = self.get_response(request)

        return response

    def _get_latest_flagged_source_url(
        self, session_key: str, source_url: str
    ) -> UserActivityFlag:
        """
        Gets the most recent flagged source URL to see whether or not the flagged
        feature was active.
        """
        source_user_activity = (
            UserActivity.objects.filter(session_key=session_key, url=source_url)
            .order_by("-timestamp")
            .first()
        )
        source_uaf = UserActivityFlag.objects.filter(
            user_activity=source_user_activity
        ).first()
        return source_uaf.is_active



# Generated by Django 4.2.4 on 2023-08-17 15:12

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_bandits.models


class Migration(migrations.Migration):
    replaces = [
        ("django_bandits", "0001_initial"),
        (
            "django_bandits",
            "0002_ucb1model_min_views_ucb1model_significance_level_and_more",
        ),
        ("django_bandits", "0003_epsilondecaymodel_min_views_and_more"),
        ("django_bandits", "0004_remove_epsilondecaymodel_beta"),
    ]

    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Bandit",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        choices=[
                            ("EG", "Epsilon Greedy"),
                            ("ED", "Epsilon Decay"),
                            ("UCB1", "Upper Confidence Bound"),
                        ],
                        default="EG",
                        max_length=200,
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                ("is_active", models.BooleanField(default=False)),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="BanditFlag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="The human/computer readable name.",
                        max_length=100,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "everyone",
                    models.BooleanField(
                        blank=True,
                        help_text="Flip this flag on (Yes) or off (No) for everyone, overriding all other settings. Leave as Unknown to use normally.",
                        null=True,
                        verbose_name="Everyone",
                    ),
                ),
                (
                    "percent",
                    models.DecimalField(
                        blank=True,
                        decimal_places=1,
                        help_text="A number between 0.0 and 99.9 to indicate a percentage of users for whom this flag will be active.",
                        max_digits=3,
                        null=True,
                        verbose_name="Percent",
                    ),
                ),
                (
                    "testing",
                    models.BooleanField(
                        default=False,
                        help_text="Allow this flag to be set for a session for user testing",
                        verbose_name="Testing",
                    ),
                ),
                (
                    "superusers",
                    models.BooleanField(
                        default=True,
                        help_text="Flag always active for superusers?",
                        verbose_name="Superusers",
                    ),
                ),
                (
                    "staff",
                    models.BooleanField(
                        default=False,
                        help_text="Flag always active for staff?",
                        verbose_name="Staff",
                    ),
                ),
                (
                    "authenticated",
                    models.BooleanField(
                        default=False,
                        help_text="Flag always active for authenticated users?",
                        verbose_name="Authenticated",
                    ),
                ),
                (
                    "languages",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Activate this flag for users with one of these languages (comma-separated list)",
                        verbose_name="Languages",
                    ),
                ),
                (
                    "rollout",
                    models.BooleanField(
                        default=False,
                        help_text="Activate roll-out mode?",
                        verbose_name="Rollout",
                    ),
                ),
                (
                    "note",
                    models.TextField(
                        blank=True,
                        help_text="Note where this Flag is used.",
                        verbose_name="Note",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        db_index=True,
                        default=django.utils.timezone.now,
                        help_text="Date when this Flag was created.",
                        verbose_name="Created",
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        help_text="Date when this Flag was last modified.",
                        verbose_name="Modified",
                    ),
                ),
                ("object_id", models.PositiveIntegerField(blank=True, null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Activate this flag for these user groups.",
                        to="auth.group",
                        verbose_name="Groups",
                    ),
                ),
                (
                    "users",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Activate this flag for these users.",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Users",
                    ),
                ),
            ],
            options={
                "verbose_name": "Flag",
                "verbose_name_plural": "Flags",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="UserActivity",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("session_key", models.CharField(max_length=40, null=True)),
                ("url", models.URLField()),
                ("target_url_visit", models.BooleanField(default=False)),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("is_staff", models.BooleanField(default=False)),
                (
                    "flags",
                    models.ManyToManyField(blank=True, to=settings.WAFFLE_FLAG_MODEL),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            bases=(django_bandits.models.URLSanitizationMixin, models.Model),
        ),
        migrations.CreateModel(
            name="UserActivityFlag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("is_active", models.BooleanField(default=False, null=True)),
                (
                    "flag",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.WAFFLE_FLAG_MODEL,
                    ),
                ),
                (
                    "user_activity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_bandits.useractivity",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UCB1Model",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_active", models.BooleanField(default=False)),
                ("c", models.FloatField(default=2.0)),
                (
                    "flag",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(class)s_set",
                        to=settings.WAFFLE_FLAG_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="FlagUrl",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("source_url", models.CharField(blank=True, max_length=200, null=True)),
                ("target_url", models.CharField(blank=True, max_length=200, null=True)),
                ("active_flag_views", models.IntegerField(default=0)),
                ("inactive_flag_views", models.IntegerField(default=0)),
                ("active_flag_conversions", models.IntegerField(default=0)),
                ("inactive_flag_conversions", models.IntegerField(default=0)),
                (
                    "flag",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.WAFFLE_FLAG_MODEL,
                    ),
                ),
            ],
            bases=(django_bandits.models.URLSanitizationMixin, models.Model),
        ),
        migrations.CreateModel(
            name="EpsilonGreedyModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_active", models.BooleanField(default=False)),
                ("epsilon", models.FloatField(default=0.1)),
                ("prob_flag", models.FloatField(default=0.5)),
                (
                    "flag",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(class)s_set",
                        to=settings.WAFFLE_FLAG_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="EpsilonDecayModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_active", models.BooleanField(default=False)),
                ("beta", models.FloatField(default=0.99)),
                (
                    "flag",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="%(class)s_set",
                        to=settings.WAFFLE_FLAG_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="BanditInstance",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "bandit",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_bandits.bandit",
                    ),
                ),
                (
                    "flag",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.WAFFLE_FLAG_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="ucb1model",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_active", True)),
                fields=("flag", "is_active"),
                name="unique_active_bandit_ucb1model",
            ),
        ),
        migrations.AddConstraint(
            model_name="epsilongreedymodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_active", True)),
                fields=("flag", "is_active"),
                name="unique_active_bandit_epsilongreedymodel",
            ),
        ),
        migrations.AddConstraint(
            model_name="epsilondecaymodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_active", True)),
                fields=("flag", "is_active"),
                name="unique_active_bandit_epsilondecaymodel",
            ),
        ),
        migrations.AddField(
            model_name="ucb1model",
            name="min_views",
            field=models.IntegerField(default=100),
        ),
        migrations.AddField(
            model_name="ucb1model",
            name="significance_level",
            field=models.FloatField(default=0.05),
        ),
        migrations.AddField(
            model_name="ucb1model",
            name="winning_arm",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="epsilondecaymodel",
            name="min_views",
            field=models.IntegerField(default=100),
        ),
        migrations.AddField(
            model_name="epsilondecaymodel",
            name="significance_level",
            field=models.FloatField(default=0.05),
        ),
        migrations.AddField(
            model_name="epsilondecaymodel",
            name="winning_arm",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="epsilongreedymodel",
            name="min_views",
            field=models.IntegerField(default=100),
        ),
        migrations.AddField(
            model_name="epsilongreedymodel",
            name="significance_level",
            field=models.FloatField(default=0.05),
        ),
        migrations.AddField(
            model_name="epsilongreedymodel",
            name="winning_arm",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.RemoveField(
            model_name="epsilondecaymodel",
            name="beta",
        ),
    ]
